#include "OgreVisibilityTerrainSceneManager.h"
#include "OgreVisibilityOptionsManager.h"
#include <OgreMath.h>
#include <OgreIteratorWrappers.h>
#include <OgreRenderSystem.h>
#include <OgreCamera.h>
#include <OgreLogManager.h>
#include <OgreStringConverter.h>
#include <OgreEntity.h>
#include <OgreSubEntity.h>


namespace Ogre {

//-----------------------------------------------------------------------
VisibilityTerrainSceneManager::VisibilityTerrainSceneManager(GtpVisibility::
															 VisibilityManager *visManager): 
mVisibilityManager(visManager), 
mShowVisualization(false),
mRenderNodesForViz(false),
mRenderNodesContentForViz(false),
mVisualizeCulledNodes(false),
mLeavePassesInQueue(0),
mDelayRenderTransparents(true),
mUseDepthPass(false),
mRenderDepthPass(false),
mUseItemBuffer(false),
//mUseItemBuffer(true),
mRenderItemBuffer(false),
mCurrentEntityId(1),
mEnableDepthWrite(true),
mSkipTransparents(false),
mSavedShadowTechnique(SHADOWTYPE_NONE),
mRenderHierarchyNodes(false)
//mRenderHierarchyNodes(true)
{
	mHierarchyInterface = new OctreeHierarchyInterface(this, mDestRenderSystem);
	
	//mDisplayNodes = true;
	//mShowBoundingBoxes = true;

	// TODO: set maxdepth to reasonable value
	mMaxDepth = 50;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::InitDepthPass()
{
	MaterialPtr depthMat = MaterialManager::getSingleton().getByName("Visibility/DepthPass");

	if (depthMat.isNull())
    {
		// Init
		depthMat = MaterialManager::getSingleton().create(
                "Visibility/DepthPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
        mDepthPass = depthMat->getTechnique(0)->getPass(0);
		mDepthPass->setColourWriteEnabled(false);
		mDepthPass->setDepthWriteEnabled(true);
		mDepthPass->setLightingEnabled(false);
	}
	else
	{
		mDepthPass = depthMat->getTechnique(0)->getPass(0);
	}
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::InitItemBufferPass()
{
	MaterialPtr itemBufferMat = MaterialManager::getSingleton().
		getByName("Visibility/ItemBufferPass");

	if (itemBufferMat.isNull())
    {
		// Init
		itemBufferMat = MaterialManager::getSingleton().create("Visibility/ItemBufferPass",
                ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);

		mItemBufferPass = itemBufferMat->getTechnique(0)->getPass(0);
		mItemBufferPass->setColourWriteEnabled(true);
		mItemBufferPass->setDepthWriteEnabled(true);
		mItemBufferPass->setLightingEnabled(true);
		//mItemBufferPass->setLightingEnabled(false);
	}
	else
	{
		mItemBufferPass = itemBufferMat->getTechnique(0)->getPass(0);
	}
	//mItemBufferPass->setAmbient(1, 1, 0);
}
//-----------------------------------------------------------------------
VisibilityTerrainSceneManager::~VisibilityTerrainSceneManager()
{
	if (mHierarchyInterface)
	{
		delete mHierarchyInterface;
		mHierarchyInterface = NULL;
	}
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::PrepareVisualization(Camera *cam)
{
	// add player camera for visualization purpose
	try 
	{
		Camera *c;
		if ((c = getCamera("PlayerCam")) != NULL)
		{
			getRenderQueue()->addRenderable(c);
		}   
    }
    catch (...)
    {
        // ignore
    }
	for (BoxList::iterator it = mBoxes.begin(); it != mBoxes.end(); ++it)
	{
		getRenderQueue()->addRenderable(*it);
	}
	if (mRenderNodesForViz || mRenderNodesContentForViz)
	{
		// change node material so it is better suited for visualization
		MaterialPtr nodeMat = MaterialManager::getSingleton().getByName("Core/NodeMaterial");
		nodeMat->setAmbient(1, 1, 0);
		nodeMat->setLightingEnabled(true);
		nodeMat->getTechnique(0)->getPass(0)->removeAllTextureUnitStates();

		for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
		{
			if (mRenderNodesForViz)
			{
				if (((*it)->numAttachedObjects() > 0) && ((*it)->numChildren() == 0)
					&& (*it)->getAttachedObject(0)->getMovableType() == "Entity")
				getRenderQueue()->addRenderable((*it));

				// addbounding boxes instead of node itself
				//(*it)->_addBoundingBoxToQueue(getRenderQueue());
			}
			if (mRenderNodesContentForViz) 
			{
				(*it)->_addToRenderQueue(cam, getRenderQueue(), false);
			}
		}
	}	
}
//-----------------------------------------------------------------------
Pass *VisibilityTerrainSceneManager::setPass(Pass* pass)
{
	// TODO: setting vertex program is not efficient
	//Pass *usedPass = ((mRenderDepthPass && !pass->hasVertexProgram()) ? mDepthPass : pass); 
	
	// set depth fill pass if we currently do not make an aabb occlusion query
	Pass *usedPass = (mRenderDepthPass && !mHierarchyInterface->IsBoundingBoxQuery() ? 
					  mDepthPass : pass);
		
	IlluminationRenderStage savedStage = mIlluminationStage; 
	
	// set illumination stage to NONE so no shadow material is used 
	// for depth pass or for occlusion query
	if (mRenderDepthPass || mHierarchyInterface->IsBoundingBoxQuery())
	{
		mIlluminationStage = IRS_NONE;
	}
	
	if (mRenderDepthPass)
	{
		// --- set vertex program of current pass so z-buffer is updated correctly
        if (pass->hasVertexProgram())
		{
			mDepthPass->setVertexProgram(pass->getVertexProgramName());

			if (mDepthPass->hasVertexProgram())
			{
				const GpuProgramPtr& prg = mDepthPass->getVertexProgram();
				// Load this program if not done already
				if (!prg->isLoaded())
					prg->load();
				// Copy params
				mDepthPass->setVertexProgramParameters(pass->getVertexProgramParameters());
			}
		}
		else if (mDepthPass->hasVertexProgram())
		{
			mDepthPass->setVertexProgram("");
		}
	}

	bool IsDepthWrite = usedPass->getDepthWriteEnabled();

	// global option which enables / disables depth writes
	if (!mEnableDepthWrite)
	{
		usedPass->setDepthWriteEnabled(false);
	}
	//else if (mIsItemBufferPass) {usedPass = mItemBufferPass;}
	
	Pass *result = SceneManager::setPass(usedPass);

	// reset depth write
	if (!mEnableDepthWrite)
	{
		usedPass->setDepthWriteEnabled(IsDepthWrite);
	}

	// reset illumination stage
	mIlluminationStage = savedStage;

	return result;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_findVisibleObjects(Camera* cam, bool onlyShadowCasters)
{
	// clear list of solid boxes (used for item buffer hierarchy node rendering)
	/*for (int i=0; i<(int)mSolidBoxes.size(); ++i)
		delete mSolidBoxes[i];

	mSolidBoxes.clear();*/

	//-- show visible scene nodes and octree bounding boxes from last frame
	if (mShowVisualization)
    {
		PrepareVisualization(cam);
	}
	else if (mRenderHierarchyNodes)
	{
		AxisAlignedBox aab;
		

		// get rendered hierarchy nodes from last frame
		GtpVisibility::HierarchyNodeList *nodeList = mHierarchyInterface->GetRenderedNodes();
		GtpVisibility::HierarchyNodeList::iterator nodeIt, nodeIt_end = nodeList->end();

		for (nodeIt = nodeList->begin(); nodeIt != nodeIt_end; ++nodeIt)		
		{
			SolidBoundingBox *solidBox = new SolidBoundingBox();
			solidBox->SetupBoundingBox(aab);

			Octree *octree = static_cast<Octree *>(*nodeIt);
			solidBox->setId(octree->getId());
			mSolidBoxes.push_back(solidBox);

			aab = octree->_getWorldAABB();
			std::stringstream d; d << "bounding box with id: " << octree->getId() << ", " << aab << "\n";
			LogManager::getSingleton().logMessage(d.str());

			getRenderQueue()->addRenderable(solidBox);	
		}
	}
	else 
	{	
		// for hierarchical culling, we interleave identification 
		// and rendering of objects in _renderVisibibleObjects

		// for the shadow pass we use only standard rendering
		// because of low occlusion
		if (mShadowTechnique == SHADOWTYPE_TEXTURE_MODULATIVE && 
			mIlluminationStage == IRS_RENDER_TO_TEXTURE)
		{
			TerrainSceneManager::_findVisibleObjects(cam, onlyShadowCasters);
		}
		// only shadow casters will be rendered in shadow texture pass
		// mHierarchyInterface->SetOnlyShadowCasters(onlyShadowCasters);
	}
	//TerrainSceneManager::_findVisibleObjects(cam, onlyShadowCasters);
	
	// -- delete lists stored for visualization
	mVisible.clear();
	mBoxes.clear();
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_renderVisibleObjects()
{
	// save ambient light to reset later
	ColourValue savedAmbient = mAmbientLight;

	// --- apply standard rendering for some cases 
    // e.g., visualization, shadow pass

	if (mShowVisualization || mRenderHierarchyNodes ||
	   (mShadowTechnique == SHADOWTYPE_TEXTURE_MODULATIVE && 
	    mIlluminationStage == IRS_RENDER_TO_TEXTURE))
	{	
		IlluminationRenderStage savedStage = mIlluminationStage; 
	
		if (mShowVisualization || mRenderHierarchyNodes) 
			// disable illumination stage because we want no shadows
			mIlluminationStage = IRS_NONE;

		// standard rendering for shadow maps because of performance
		TerrainSceneManager::_renderVisibleObjects();

		mIlluminationStage = savedStage;
	}
	else // -- the hierarchical culling algorithm
	{
		// don't render backgrounds for item buffer
		if (mUseItemBuffer)
		{
			clearSpecialCaseRenderQueues();
			getRenderQueue()->clear();
		}

		// the objects of different layers (e.g., background, scene, 
		// overlay) must be identified and rendered one after another

		//-- render all early skies
		clearSpecialCaseRenderQueues();
		addSpecialCaseRenderQueue(RENDER_QUEUE_BACKGROUND);
		addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_EARLY);
		setSpecialCaseRenderQueueMode(SceneManager::SCRQM_INCLUDE);

		TerrainSceneManager::_renderVisibleObjects();
		

#ifdef GTP_VISIBILITY_MODIFIED_OGRE
		// delete previously rendered contenbt
		_deleteRenderedQueueGroups();
#endif

		//-- prepare queue for visible objects (i.e., all but overlay and skies late)
		clearSpecialCaseRenderQueues();
		addSpecialCaseRenderQueue(RENDER_QUEUE_SKIES_LATE);
		addSpecialCaseRenderQueue(RENDER_QUEUE_OVERLAY);
		setSpecialCaseRenderQueueMode(SceneManager::SCRQM_EXCLUDE);


		// set all necessary parameters for 
		// hierarchical visibility culling and rendering
		InitVisibilityCulling(mCameraInProgress);

		/** 
		* the hierarchical culling algorithm
		* if we use a depth pass: will just find objects and update depth buffer
		* for "delayed" rendering: will render some passes afterwards, e.g., transparents
		**/
		
		mVisibilityManager->ApplyVisibilityCulling();

		// delete remaining renderables from queue (all not in mLeavePassesInQueue)
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
		_deleteRenderedQueueGroups(mLeavePassesInQueue);
#endif

		//-- reset parameters
		mRenderDepthPass = false;
		mRenderItemBuffer = false;
		mSkipTransparents = false;
		mLeavePassesInQueue = 0;
		mShadowTechnique = mSavedShadowTechnique;
		

		// add visible nodes found by the visibility culling algorithm
		if (mUseDepthPass)
		{
			for (NodeList::iterator it = mVisible.begin(); it != mVisible.end(); ++it)
			{
				(*it)->_addToRenderQueue(mCameraInProgress, getRenderQueue(), false);
			}
		}
		
		//-- now we can render all remaining queue objects
		// for depth pass, transparents, overlay 
		clearSpecialCaseRenderQueues();
		TerrainSceneManager::_renderVisibleObjects();
	}
		
	// set the new render level index afterwards => new level in the next frame 
	int levelIdx = TerrainRenderable::getCurrentRenderLevelIndex() + 1;
	TerrainRenderable::setCurrentRenderLevelIndex(levelIdx);

	// reset ambient light
	setAmbientLight(savedAmbient);

	getRenderQueue()->clear();
	//WriteLog(); // write out stats
}

//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::_updateSceneGraph(Camera* cam)
{
	mVisibilityManager->GetCullingManager()->SetHierarchyInterface(mHierarchyInterface);
	mHierarchyInterface->SetRenderSystem(mDestRenderSystem);

	TerrainSceneManager::_updateSceneGraph(cam);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::setOption(const String & key, const void * val)
{
	if (key == "UseDepthPass")
	{
		mUseDepthPass = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "PrepareVisualization")
	{
		mShowVisualization = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesForViz")
	{
		mRenderNodesForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderNodesContentForViz")
	{
		mRenderNodesContentForViz = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyBoxEnabled")
	{
		mSkyBoxEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyPlaneEnabled")
	{
		mSkyPlaneEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "SkyDomeEnabled")
	{
		mSkyDomeEnabled = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "VisualizeCulledNodes")
	{
		mVisualizeCulledNodes = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "DelayRenderTransparents")
	{
		mDelayRenderTransparents = (*static_cast<const bool *>(val));
		return true;
	}
	// notifiy that frame has ended so terrain render level can be reset for correct
	// terrain rendering
	if (key == "TerrainLevelIdx")
	{
		TerrainRenderable::setCurrentRenderLevelIndex((*static_cast<const int *>(val)));
		return true;
	}
	if (key == "DepthWrite")
	{
		mEnableDepthWrite = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "UseItemBuffer")
	{
		mUseItemBuffer = (*static_cast<const bool *>(val));
		return true;
	}
	if (key == "RenderHierarchyNodes")
	{
		mRenderHierarchyNodes = (*static_cast<const bool *>(val));
	}
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		setOption(key, val) || TerrainSceneManager::setOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOption(const String & key, void *val)
{
	if (key == "NumHierarchyNodes")
	{
		* static_cast<unsigned int *>(val) = (unsigned int)mNumOctreeNodes;
		return true;
	}
	
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOption(key, val) && TerrainSceneManager::getOption(key, val);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOptionValues(const String & key, 
													StringVector &refValueList)
{
	return TerrainSceneManager::getOptionValues( key, refValueList);
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::getOptionKeys(StringVector & refKeys)
{
	return VisibilityOptionsManager(mVisibilityManager, mHierarchyInterface).
		getOptionKeys(refKeys) || TerrainSceneManager::getOptionKeys(refKeys);
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::setVisibilityManager(GtpVisibility::
														 VisibilityManager *visManager)
{
	mVisibilityManager = visManager;
}
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *VisibilityTerrainSceneManager::getVisibilityManager( void )
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::WriteLog()
{
	std::stringstream d;

	d << "Depth pass: " << StringConverter::toString(mUseDepthPass) << ", "
	  << "Delay transparents: " << StringConverter::toString(mDelayRenderTransparents) << ", "
	  << "Use optimization: " << StringConverter::toString(mHierarchyInterface->GetTestGeometryForVisibleLeaves()) << ", "
	  << "Algorithm type: " << mVisibilityManager->GetCullingManagerType() << ", "
	  << "Hierarchy nodes: " << mNumOctreeNodes << ", " 
	  << "Traversed nodes: " << mHierarchyInterface->GetNumTraversedNodes() << ", "
	  << "Rendered nodes: " << mHierarchyInterface->GetNumRenderedNodes() << ", "
	  << "Query culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumQueryCulledNodes() << ", "
	  << "Frustum culled nodes: " << mVisibilityManager->GetCullingManager()->GetNumFrustumCulledNodes() << ", "
      << "Queries issued: " << mVisibilityManager->GetCullingManager()->GetNumQueriesIssued() << "\n";

	LogManager::getSingleton().logMessage(d.str());
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::renderObjects(
	const RenderPriorityGroup::TransparentRenderablePassList& objs, 
    bool doLightIteration, const LightList* manualLightList)
{
	// for correct rendering, transparents must be rendered after hierarchical culling
	if (!mSkipTransparents)
	{
		OctreeSceneManager::renderObjects(objs, doLightIteration, manualLightList);
	}
}
//-----------------------------------------------------------------------
bool VisibilityTerrainSceneManager::validatePassForRendering(Pass* pass)
{
	// skip all but first pass if we are doing the depth pass
	if ((mRenderDepthPass || mRenderItemBuffer) && pass->getIndex() > 0)
	{
		return false;
	}
	return SceneManager::validatePassForRendering(pass);
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::renderQueueGroupObjects(RenderQueueGroup* pGroup)
{
	if (!mRenderItemBuffer)
	{
		TerrainSceneManager::renderQueueGroupObjects(pGroup);
		return;
	}

	// --- item buffer

    // Iterate through priorities
    RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();

	while (groupIt.hasMoreElements())
    {
		RenderItemBuffer(groupIt.getNext());
	}
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::RenderItemBuffer(RenderPriorityGroup* pGroup)
{
	// Do solids
	RenderPriorityGroup::SolidRenderablePassMap solidObjs = pGroup->_getSolidPasses();

	// ----- SOLIDS LOOP -----
	RenderPriorityGroup::SolidRenderablePassMap::const_iterator ipass, ipassend;
	ipassend = solidObjs.end();

	for (ipass = solidObjs.begin(); ipass != ipassend; ++ipass)
	{
		// Fast bypass if this group is now empty
		if (ipass->second->empty()) 
			continue;

		// Render only first pass
		if (ipass->first->getIndex() > 0)
			continue;

		RenderPriorityGroup::RenderableList* rendList = ipass->second;
		
		RenderPriorityGroup::RenderableList::const_iterator irend, irendend;
		irendend = rendList->end();
			
		for (irend = rendList->begin(); irend != irendend; ++irend)
		{
			std::stringstream d; d << "itembuffer, pass name: " << 
				ipass->first->getParent()->getParent()->getName();
				
			LogManager::getSingleton().logMessage(d.str());
			
			RenderSingleObjectForItemBuffer(*irend, ipass->first);
		}
	}

	// ----- TRANSPARENT LOOP: must be handled differently 
	// TODO: HOW TO HANDLE OCCLUDED OBJECTS ????
	RenderPriorityGroup::TransparentRenderablePassList
		transpObjs = pGroup->_getTransparentPasses();
	RenderPriorityGroup::TransparentRenderablePassList::const_iterator itrans, itransend;

	itransend = transpObjs.end();
	for (itrans = transpObjs.begin(); itrans != itransend; ++itrans)
	{
		// like for solids, render only first pass
		if (itrans->pass->getIndex() == 0)
		{	
			RenderSingleObjectForItemBuffer(itrans->renderable, itrans->pass);
		}
	} 
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::RenderSingleObjectForItemBuffer(Renderable *rend, Pass *pass)
{
	static LightList nullLightList;
	
	int col[4];
	
	// -- create color code out of object id
	col[0] = (rend->getId() >> 16) & 255;
	col[1] = (rend->getId() >> 8) & 255;
	col[2] = rend->getId() & 255;
//	col[3] = 255;

	//mDestRenderSystem->setColour(col[0], col[1], col[2], col[3]);
    
	mItemBufferPass->setAmbient(ColourValue(col[0] / 255.0f,
										    col[1] / 255.0f, 
											col[2] / 255.0f, 1));

	// set vertex program of current pass
	if (pass->hasVertexProgram())
	{
		mItemBufferPass->setVertexProgram(pass->getVertexProgramName());

		if (mItemBufferPass->hasVertexProgram())
		{
			const GpuProgramPtr& prg = mItemBufferPass->getVertexProgram();
			// Load this program if not done already
			if (!prg->isLoaded())
				prg->load();
			// Copy params
			mItemBufferPass->setVertexProgramParameters(pass->getVertexProgramParameters());
		}
	}
	else if (mItemBufferPass->hasVertexProgram())
	{
		mItemBufferPass->setVertexProgram("");
	}

	Pass *usedPass = setPass(mItemBufferPass); 


	// Render a single object, this will set up auto params if required
	renderSingleObject(rend, usedPass, false, &nullLightList);
}
//-----------------------------------------------------------------------
GtpVisibility::VisibilityManager *VisibilityTerrainSceneManager::GetVisibilityManager()
{
	return mVisibilityManager;
}
//-----------------------------------------------------------------------
Entity* VisibilityTerrainSceneManager::createEntity(const String& entityName, 
													const String& meshName)
{
	Entity *ent = SceneManager::createEntity(entityName, meshName);

	for (int i = 0; i < (int)ent->getNumSubEntities(); ++i)
	{
		ent->getSubEntity(i)->setId(mCurrentEntityId);
	}

	// increase counter of entity id values
	++ mCurrentEntityId;

	return ent;
}
//-----------------------------------------------------------------------
void VisibilityTerrainSceneManager::InitVisibilityCulling(Camera *cam)
{
	InitDepthPass(); 	  // create material for depth pass
	InitItemBufferPass(); // create material for item buffer pass

	// reset culling manager stats
	mVisibilityManager->GetCullingManager()->InitFrame(mVisualizeCulledNodes);

	// save shadow technique. It will be reset after hierarchical culling
	mSavedShadowTechnique = mShadowTechnique;

	// render standard solids without shadows during hierarchical culling pass
	if ((mShadowTechnique == SHADOWTYPE_STENCIL_ADDITIVE) ||
	    (mShadowTechnique == SHADOWTYPE_STENCIL_MODULATIVE))
	{
	
		mShadowTechnique = SHADOWTYPE_NONE;
	}

	// set depth pass flag before rendering
	mRenderDepthPass = mUseDepthPass;

	// item buffer needs full ambient lighting to use item colors as unique id
	if (mUseItemBuffer) 
	{
		mRenderItemBuffer = true;
		setAmbientLight(ColourValue(1,1,1,1));
	}


	// set passes which should be stored in render queue 
	// for rendering AFTER hierarchical culling, i.e., passes which need 
	// a special rendering order
	mLeavePassesInQueue = 0;

	if (!mUseDepthPass || !mUseItemBuffer)
	{
		if (mShadowTechnique == SHADOWTYPE_STENCIL_ADDITIVE)
		{
			// TODO: remove this pass because it should be processed during hierarchical culling
			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_NOSHADOW;

			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_DECAL;
			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_DIFFUSE_SPECULAR;
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;

			// just render ambient stuff
			mIlluminationStage = IRS_AMBIENT;
		}
	
		if (mShadowTechnique == SHADOWTYPE_STENCIL_MODULATIVE)
		{
			mLeavePassesInQueue |= RenderPriorityGroup::SOLID_PASSES_NOSHADOW;
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;
		}
	
		// transparents should be rendered after hierarchical culling to 
		// provide front-to-back ordering
		if (mDelayRenderTransparents)
		{
			mLeavePassesInQueue |= RenderPriorityGroup::TRANSPARENT_PASSES;
		}
	}

	// skip rendering transparents in the hierarchical culling
	// (because they will be rendered afterwards)
	mSkipTransparents = mUseDepthPass || 
		(mLeavePassesInQueue & RenderPriorityGroup::TRANSPARENT_PASSES);

	// -- initialise interface for rendering traversal of the hierarchy
	mHierarchyInterface->SetHierarchyRoot(mOctree);
	
	// possible two cameras (one for culling, one for rendering)
	mHierarchyInterface->InitTraversal(mCameraInProgress, 
							mCullCamera ? getCamera("CullCamera") : NULL,
							mLeavePassesInQueue);
		
	//std::stringstream d; d << "leave passes in queue: " << mLeavePassesInQueue;LogManager::getSingleton().logMessage(d.str());
}
//-----------------------------------------------------------------------
OctreeHierarchyInterface *VisibilityTerrainSceneManager::GetHierarchyInterface()
{
	return mHierarchyInterface;
}
//-----------------------------------------------------------------------
/*void VisibilityTerrainSceneManager::renderBasicQueueGroupObjects(RenderQueueGroup* pGroup)
{
    // Basic render loop: Iterate through priorities
    RenderQueueGroup::PriorityMapIterator groupIt = pGroup->getIterator();

    while (groupIt.hasMoreElements())
    {
        RenderPriorityGroup* pPriorityGrp = groupIt.getNext();

        // Sort the queue first
        pPriorityGrp->sort(mCameraInProgress);

		// Do solids
		// TODO: render other solid passes for shadows
        renderObjects(pPriorityGrp->_getSolidPassesNoShadows(), true);

		// do solid passes no shadows if addititive stencil shadows
		if (mSavedShadowTechnique == SHADOWTYPE_STENCIL_ADDITIVE)
			renderObjects(pPriorityGrp->_getSolidPassesNoShadows(), true);
		
        // Do transparents
        renderObjects(pPriorityGrp->_getTransparentPasses(), true);


    }// for each priority
}
*/
} // namespace Ogre
